package com.solacesystems.cloudfoundry.demos.intro.controller;

import java.util.concurrent.atomic.AtomicInteger;

import javax.annotation.PostConstruct;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.json.JSONArray;
import org.json.JSONObject;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RestController;

import com.solacesystems.cloudfoundry.demos.intro.model.SimpleMessage;
import com.solacesystems.cloudfoundry.demos.intro.model.SimpleSubscription;
import com.solacesystems.common.util.trace.Trace;
import com.solacesystems.jcsmp.BytesXMLMessage;
import com.solacesystems.jcsmp.JCSMPException;
import com.solacesystems.jcsmp.JCSMPFactory;
import com.solacesystems.jcsmp.JCSMPProperties;
import com.solacesystems.jcsmp.JCSMPSession;
import com.solacesystems.jcsmp.JCSMPStreamingPublishEventHandler;
import com.solacesystems.jcsmp.TextMessage;
import com.solacesystems.jcsmp.Topic;
import com.solacesystems.jcsmp.XMLMessageConsumer;
import com.solacesystems.jcsmp.XMLMessageListener;
import com.solacesystems.jcsmp.XMLMessageProducer;

@RestController
public class SolaceController {

	private static final Log trace = LogFactory.getLog(SolaceController.class);

	JCSMPSession session;
	XMLMessageProducer producer;
	TextMessage lastReceivedMessage;
	
	// Stats
	AtomicInteger numMessagesReceived = new AtomicInteger();
	AtomicInteger numMessagesSent = new AtomicInteger();
	
	class SimplePublisherEventHandler implements JCSMPStreamingPublishEventHandler {
		@Override
		public void responseReceived(String messageID) {
			trace.debug("Producer received response for msg: " + messageID);
		}

		@Override
		public void handleError(String messageID, JCSMPException e,
				long timestamp) {
			trace.error("Producer received error for msg: "+ messageID + " - " + timestamp, e);
		}

	};
	
	class SimpleMessageListener implements XMLMessageListener {
		
		@Override
		public void onReceive(BytesXMLMessage receivedMessage) {

			numMessagesReceived.incrementAndGet();
			
			if (receivedMessage instanceof TextMessage) {
				lastReceivedMessage = (TextMessage)receivedMessage;
			} else {
				trace.error("Received message that was not a TextMessage: " + receivedMessage.dump());
			}
		}

		@Override
		public void onException(JCSMPException e) {
			trace.error("Consumer received exception: %s%n", e);
		}
	}

	@PostConstruct
	public void init() {
		
		// Connect to Solace
		System.out.println("************* Init Called ************");
		
		String vcapServices = System.getenv("VCAP_SERVICES");
		trace.error(vcapServices);

		// Must parse the Solace HOST from VCAP Services
		if (vcapServices == null || vcapServices.equals("") || vcapServices.equals("{}")) {
			trace.error("Did not find user provided service. Aborting conenction");
			return;
		}
		
		JSONObject vcapServicesJson = new JSONObject(vcapServices);
		JSONArray userProvidedServicesArray = vcapServicesJson.getJSONArray("user-provided");
		JSONObject solaceCredentials = null;
		for (int i = 0; i < userProvidedServicesArray.length(); i++) {
			JSONObject currService = userProvidedServicesArray.getJSONObject(i);
			if (currService.getString("name").equals("solace-user-provided-messaging")) {
				solaceCredentials = currService.getJSONObject("credentials");
			}
		}
		
		if (solaceCredentials == null) {
			trace.error("Did not find Solace messaging service named \"solace-user-provided-messaging\"");
			return;
		} else {
			trace.warn("Solace Credentials: " + solaceCredentials.toString());
		}
		 
		System.out.println("Solace client initializing...");
		final JCSMPProperties properties = new JCSMPProperties();
		properties.setProperty(JCSMPProperties.HOST, solaceCredentials.getString("smfUri"));
		properties.setProperty(JCSMPProperties.VPN_NAME, solaceCredentials.getString("msgVpn"));
		properties.setProperty(JCSMPProperties.USERNAME, solaceCredentials.getString("username"));
		properties.setProperty(JCSMPProperties.PASSWORD, solaceCredentials.getString("password"));
		
		try {
			session = JCSMPFactory.onlyInstance().createSession(properties);
			session.connect();
			
			final XMLMessageConsumer cons = session.getMessageConsumer(new SimpleMessageListener());
			cons.start();
			
			producer = session.getMessageProducer(new SimplePublisherEventHandler());
			
			
			System.out.println("************* Solace initialized correctly!! ************");
		} catch (Exception e) {
			Trace.error("Error connecting and setting up session.", e);
		}
	}
	
	@RequestMapping(value = "/message", method = RequestMethod.POST)
	public ResponseEntity<String> sendMessage(@RequestBody SimpleMessage message) {

		trace.info("Sending message on topic: " + message.getTopic() + " with body: " + message.getBody());
		
		final Topic topic = JCSMPFactory.onlyInstance().createTopic(message.getTopic());
		TextMessage msg = JCSMPFactory.onlyInstance().createMessage(TextMessage.class);
		msg.setText(message.getBody());
		try {
			producer.send(msg,topic);
			numMessagesSent.incrementAndGet();
			
		} catch (JCSMPException e) {
			trace.error("Service Creation failed.", e);
			return new ResponseEntity<>("{'description': '" + e.getMessage() + "'}", HttpStatus.BAD_REQUEST);
		}
		return new ResponseEntity<>("{}", HttpStatus.OK);
	}
	
	@RequestMapping(value = "/message", method = RequestMethod.GET)
	public ResponseEntity<SimpleMessage> getLastMessageReceived() {

		// Return the last received message if it exists.
		SimpleMessage receivedMessage = new SimpleMessage();
		
		if (lastReceivedMessage != null) {
			receivedMessage.setTopic(lastReceivedMessage.getDestination().getName());
			receivedMessage.setBody(lastReceivedMessage.getText());
		} 		
		return new ResponseEntity<>(receivedMessage, HttpStatus.OK);
	}
	
	@RequestMapping(value = "/subscription", method = RequestMethod.POST)
	public ResponseEntity<String> addSubscription(@RequestBody SimpleSubscription subscription) {
		final Topic topic = JCSMPFactory.onlyInstance().createTopic(subscription.getSubscription());
		try {
			boolean waitForConfirm = true;
			session.addSubscription(topic, waitForConfirm);
		} catch (JCSMPException e) {
			trace.error("Service Creation failed.", e);
			return new ResponseEntity<>("{'description': '" + e.getMessage() + "'}", HttpStatus.BAD_REQUEST);
		}
		return new ResponseEntity<>("{}", HttpStatus.OK);
	}
	
	@RequestMapping(value = "/subscription", method = RequestMethod.DELETE)
	public ResponseEntity<String> deleteSubscription(@RequestBody SimpleSubscription subscription) {
		final Topic topic = JCSMPFactory.onlyInstance().createTopic(subscription.getSubscription());
		try {
			boolean waitForConfirm = true;
			session.removeSubscription(topic, waitForConfirm);
		} catch (JCSMPException e) {
			trace.error("Service Creation failed.", e);
			return new ResponseEntity<>("{'description': '" + e.getMessage() + "'}", HttpStatus.BAD_REQUEST);
		}
		return new ResponseEntity<>("{}", HttpStatus.OK);
	}	
	
	
	@RequestMapping(value = "/status", method = RequestMethod.GET)
	public ResponseEntity<String> getStatus() {

		trace.info("Received request for getStatus");
		
		JSONObject statusJson = new JSONObject();
		statusJson.put("numMsgsSent", numMessagesSent.get());
		statusJson.put("numMsgsReceived", numMessagesReceived.get());
		return new ResponseEntity<>(statusJson.toString(), HttpStatus.OK);
	}
	
	
	@RequestMapping(value = "/status", method = RequestMethod.DELETE)
	public ResponseEntity<String> resetStats() {

		numMessagesReceived.set(0);
		numMessagesSent.set(0);
		return new ResponseEntity<>("{}", HttpStatus.OK);
	}

}
